#include <iomanip>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <sys/stat.h>   // for mkdir
#include <sys/types.h>
#include <assert.h>
#include <stdlib.h>     // for (s)random

#include <list>
#include <iostream>
#include <algorithm>

#include "imms.h"
#include "strmanip.h"
#include "playlist.h"
#include "md5digest.h"

using std::endl;
using std::cerr;
using std::setprecision;
using std::list;


//////////////////////////////////////////////
// Constants

#define     FIRST_SKIP_RATE         -6
#define     CONS_SKIP_RATE          -4
#define     FIRST_NON_SKIP_RATE     5
#define     CONS_NON_SKIP_RATE      1
#define     JUMPED_TO_FINNISHED     7
#define     NO_XIDLE_BONUS          1
#define     INTERACTIVE_BONUS       2
#define     JUMPED_FROM             -1
#define     JUMPED_TO_SKIPPED       1

#define     SAMPLE_SIZE             20
#define     MIN_SAMPLE_SIZE         10
#define     DISPERSION_FACTOR       3.7
#define     MAX_TIME                20*DAY
#define     MAX_RATING              150
#define     MIN_RATING              65
#define     CORRELATION_FACTOR      3

#define     TERM_WIDTH              80
#define     HOUR                    60
#define     DAY                     (24*HOUR)


//////////////////////////////////////////////

// Random
int imms_random(int max)
{
    int rand_num;
    static struct random_data rand_data;
    static char rand_state[256];
    static bool initialized = false;
    if (!initialized)
    {
        initstate_r(time(0), rand_state, sizeof(rand_state), &rand_data);
        initialized = true;
    }
    random_r(&rand_data, &rand_num);
    double cof = rand_num / (RAND_MAX + 1.0);
    return (int)(max * cof);
}

// Imms

Imms::Imms() : timehold(MAX_TIME / 2), last_handpicked(-1)
{
    last_skipped = last_jumped = false;
    have_candidates = link = next_timehold = 0;

    string dotimms = getenv("HOME");
    mkdir(dotimms.append("/.imms").c_str(), 0700);

    fout.open(dotimms.append("/imms.log").c_str(),
            ofstream::out | ofstream::app);

    time_t t = time(0);
    fout << endl << endl << ctime(&t) << setprecision(3);

    immsdb.reset(new ImmsDb());
}

void Imms::setup(const char* _email, bool _use_xidle)
{
    email = _email;
    use_xidle = _use_xidle;
}

void Imms::pump()
{
    xidle.query();
}

void Imms::playlist_changed()
{
    timehold = MAX_TIME / 2;
}

bool Imms::add_candidate(int playlist_num, string path, bool urgent)
{
    if (!have_candidates)
        max_composite = -INT_MAX, min_composite = INT_MAX;

    ++have_candidates;

    if (urgent && have_candidates >= MIN_SAMPLE_SIZE)
        return false;
    if (have_candidates >= SAMPLE_SIZE)
        return false;

    if (candidates.count(playlist_num))
        return true;

    path_simplifyer(path);

    if (!fetch_song_info(path))
    {
        --have_candidates;
        return true;
    }

    int composite_rating = rating + link * CORRELATION_FACTOR;

    assert(composite_rating < 200);

    next_timehold += last_played;

    if (last_played > timehold)
    {
        int timebonus = 10 * (last_played - timehold) / (timehold + 1);
        timebonus = timebonus > 20 ? 20 : timebonus;
        assert(timebonus < 21);
        composite_rating += timebonus;
    }
    else if (last_played > HOUR && last_played + DAY > timehold)
    {
        int scale = (DAY - timehold + last_played) / DAY;
        assert(scale <= 1);
        composite_rating *= scale;
    }
    else
        composite_rating = 0;

    assert(composite_rating < 300);

    if (composite_rating > max_composite)
        max_composite = composite_rating;
    if (composite_rating && composite_rating < min_composite)
        min_composite = composite_rating;

    candidates[playlist_num] = composite_rating;

    return true;
}

int Imms::select_next()
{
    if (candidates.empty())
        return 0;

    int size = candidates.size();
    timehold = next_timehold / size;
    next_timehold = 0;

    int next_song;
    if ((next_song = select_next_helper()) > 0)
        return next_song;

    CandidateMap::iterator i;
    int index = imms_random(size);
    for (i = candidates.begin(); index--; ++i) { }

    return i->first;
}

int Imms::select_next_helper()
{
    CandidateMap::iterator i;

    if (min_composite == max_composite)
        return -1;
    
    int total = 0;
    for (i = candidates.begin(); i != candidates.end(); i++)
    {
        if (i->second)
            i->second = ROUND(pow(((double)(i->second - min_composite))
                        / DISPERSION_FACTOR, DISPERSION_FACTOR));
        total += i->second;
    }

    if (!total)
        return -1;

#ifdef DEBUG
    cerr << string(TERM_WIDTH - 15, '-') << endl;
    cerr << " >> [" << min_composite << "-" << max_composite << "] ";
#endif

    int next_song = -1, weight_index = imms_random(total);

    for (i = candidates.begin(); i != candidates.end(); i++)
    { 
#ifdef DEBUG
        cerr << "{" << i->second << "} ";
#endif
        weight_index -= i->second;
        if (weight_index < 0)
        {
            next_song = i->first;
#ifdef DEBUG
            weight_index = INT_MAX;
            cerr << "<*= ";
#else
            break;
#endif
        }
    }

#ifdef DEBUG
    cerr << endl;
#endif

    return next_song;
}

void Imms::start_song(string path)
{
    next_timehold = have_candidates = 0;
    candidates.clear();

    path_simplifyer(path);

    fetch_song_info(path);
    current_rating = rating;

    immsdb->set_last(time(0));
    immsdb->save_state();

    xidle.reset();

    print_song_info(path);
}

void Imms::print_song_info(const string &path)
{
    fout << string(TERM_WIDTH - 15, '-') << endl;

    fout << "[Play] [";

    if (path.length() > TERM_WIDTH - 8)
        fout << "..." << path.substr(path.length() - TERM_WIDTH + 13);
    else
        fout << path;

    fout << "]\n  [Rating: " << rating;
    if (link)
        fout << setiosflags(std::ios::showpos) << link
            << resetiosflags(std::ios::showpos) << "!";

    fout << "] [Time: " << strtime(last_played) << "/"
        << strtime(timehold) << "] ";

    if (!identified) 
        fout << "[Unknown] ";

    fout.flush();
}

void Imms::end_song(bool at_the_end, bool jumped, bool bad)
{
    int mod;

    if (at_the_end)
    {
        if (last_jumped)
            mod = JUMPED_TO_FINNISHED;
        else if (last_skipped)
            mod = FIRST_NON_SKIP_RATE;
        else
            mod = CONS_NON_SKIP_RATE;

        if (use_xidle && xidle.is_active())
            mod += INTERACTIVE_BONUS;
        else if (!use_xidle)
            mod += NO_XIDLE_BONUS;

        last_skipped = jumped = false;
    }
    else
    {
        if (last_jumped && !jumped)
            mod = JUMPED_TO_SKIPPED;
        else if (jumped)
            mod = JUMPED_FROM;
        else if (last_skipped)
            mod = CONS_SKIP_RATE;
        else
            mod = FIRST_SKIP_RATE;

        last_skipped = true;
    }

    if (bad)
        mod = 0;

    immsdb->load_state();

    if (mod > CONS_NON_SKIP_RATE + INTERACTIVE_BONUS)
        last_handpicked = immsdb->get_sid();

    fout << (jumped ? "[Jumped] " : "");
    fout << (!jumped && last_skipped ? "[Skipped] " : "");
    fout << "[Delta " << setiosflags(std::ios::showpos) << mod <<
        resetiosflags (std::ios::showpos) << "] ";
    fout << (unrated ? "[New] " : "");
    fout << endl;

    last_jumped = jumped;

    if (abs(mod) > 1)
        immsdb->add_recent(mod);

    int new_rating = current_rating + mod;
    if (new_rating > MAX_RATING)
        new_rating = MAX_RATING;
    else if (new_rating < MIN_RATING)
        new_rating = MIN_RATING;

    immsdb->set_rating(new_rating);
}

bool Imms::fetch_song_info(const string &path)
{
    if (access(path.c_str(), R_OK))
        return false;

    struct stat statbuf;
    stat(path.c_str(), &statbuf);

    if (immsdb->identify(path, statbuf.st_mtime) < 0)
    {
        if (immsdb->identify(path, statbuf.st_mtime,
                Md5Digest::digest_file(path)) < 0)
            return false;
    }

    songinfo.link(path);
    rating = immsdb->get_rating();

    unrated = false;
    if (rating < 0)
    {
        unrated = true;
        rating = songinfo.get_rating(email);
        immsdb->set_rating(rating);
    }

    StringPair info = immsdb->get_info();
    artist = info.first;
    title = info.second;

    if (artist != "" && title != "")
    {
        identified = true;
    }
    else
    {
        parse_song_info(path);
        if (identified)
            immsdb->set_title(title);
    }

    link = 0;
    if (last_handpicked != -1)
        link = immsdb->correlate(last_handpicked);

    if (link > 15)
        link = 15;

#ifdef DEBUG
    cerr << "path:\t" << path << endl;
    //cerr << "artist:\t" << artist << (identified ? " *" : "") << endl;
    //cerr << "title:\t" << title << (identified ? " *" : "") << endl;
#endif

    last_played = (time(0) - immsdb->get_last()) / 60;

    if (last_played > MAX_TIME)
        last_played = MAX_TIME;

    return true;
}

void Imms::parse_song_info(const string &path)
{

#define REMIXCLUES "rmx|mix|[^a-z]version|edit|original|remaster|" \
    "cut|instrumental|extended"
#define BADARTIST "artist|^va$|various|collection|^misc"

    //////////////////////////////////////////////
    // Initialize

    identified = false;
    bool parser_confident, artist_confirmed = false;

    string tag_artist = artist = string_normalize(songinfo.get_artist());

    StringPair fm = get_simplified_filename_mask(path);
    fm.second = string_normalize(fm.second);

    list<string> file_parts;
    parser_confident = imms_magic_parse_filename(file_parts, fm.first);

    // Can't normalize before we pass it to magic parse
    fm.first = string_normalize(fm.first);

    list<string> path_parts;
    imms_magic_parse_path(path_parts, path_get_dirname(path));

#ifdef DEBUG
#if 1
    cerr << "path parts: " << endl;
    for (list<string>::iterator i = path_parts.begin();
            i != path_parts.end(); i++)
    {
        cerr << " >> " << *i << endl;
    }
    cerr << "name parts: " << endl;
    for (list<string>::iterator i = file_parts.begin();
            i != file_parts.end(); i++)
    {
        cerr << " >> " << *i << endl;
    }
#endif
#endif

    //////////////////////////////////////////////
    // Try to identify the artist
    
    bool bad_tag_artist = tag_artist == "" || Regexx(tag_artist, BADARTIST);

    // See if we can recognize any of the path parts
    if (file_parts.size() > 1)
    {
        for (list<string>::iterator i = file_parts.begin();
                !artist_confirmed && i != file_parts.end(); i++)
        {
            if (immsdb->check_artist(*i)
                    || (!bad_tag_artist && string_like(*i, tag_artist, 4)))
            {
                artist_confirmed = true;
                artist = *i;
            }
        }
    } 

    // See if any of the directories look familiar
    string outermost = "";
    for (list<string>::iterator i = path_parts.begin();
            !artist_confirmed && i != path_parts.end(); i++)
    {
        if (Regexx(*i, BADARTIST))
            continue;
        if (artist_confirmed =
                (immsdb->check_artist(*i) || string_like(*i, artist, 4)))
            artist = *i;
        // And while we are at it find the uppermost related directory
        else if (outermost == "" && fm.first.find(*i) != string::npos)
            outermost = *i;
    }

    // The path part that is also the outermost directory might be the artist
    if (!artist_confirmed && outermost != "" && fm.second.find(outermost) < 2)
    {
        artist = outermost;
        artist_confirmed = true;
    }

    if (!artist_confirmed && parser_confident && file_parts.size() > 1)
    {
        artist = file_parts.front();
        artist_confirmed = true;
    }

    // Give the tag artist an opportunity to override the guessed artist
    if (!bad_tag_artist && !string_like(tag_artist, artist, 4))
    {
        if (immsdb->check_artist(tag_artist)
                || fm.first.find(tag_artist) != string::npos)
        {
            artist = tag_artist;
            artist_confirmed = true;
        }
    }

    // If we still don't know what to do - give up
    if (!artist_confirmed)
        return;

    // Erase the first occurrence of the artist
    list<string>::iterator i = find(file_parts.begin(), file_parts.end(), artist);
    if (i != file_parts.end())
        file_parts.erase(i);

    // By this point the artist had to have been confirmed
    immsdb->set_artist(artist);

    //////////////////////////////////////////////
    // Try (not very hard) to identify the album

    string album = album_filter(songinfo.get_album());
    string directory = album_filter(path_parts.back());
    if (album == "" || Regexx(directory, album))
        album = directory;

    //////////////////////////////////////////////
    // Try to identify the title

    string tag_title = string_tolower(songinfo.get_title());
    title = title_filter(tag_title);
    // If we know the title and were only missing the artist
    if (title != "" && (identified = immsdb->check_title(title)))
        return;

    if (identified = Regexx(title, album + ".*(" REMIXCLUES ")"))
        title = album;
    else if (identified = Regexx(fm.first, album + "$"))
        title = album;

    // Can recognize the title by matching filename parts to the database?
    for (list<string>::reverse_iterator i = file_parts.rbegin();
            !identified && i != file_parts.rend(); i++)
    {
        if (immsdb->check_title(*i) || string_like(*i, title, 4))
        {
            identified = true;
            title = *i;
        }
        else if (identified = Regexx(*i, album + ".*(" REMIXCLUES ")"))
            title = album;
    }

    if (identified)
        return;

    // See if we can trust the tag title enough to just default to it
    if (title != "" && !Regexx(title, "(" REMIXCLUES "|^track$|^title$)")
            && tag_artist != "" && !Regexx(tag_artist, BADARTIST)
            && string_like(tag_title, title, 6))
    {
        identified = true;
        return;
    }

    // Filter things that are probably not the title
    for (list<string>::iterator i = file_parts.begin();
            i != file_parts.end(); i++)
    {
        if (Regexx(*i, "(" REMIXCLUES "|^track$|^title$)"))
            *i = "###";
    }

    file_parts.remove("###");

    // Default to the right most non-suspicious entry as the title
    if (parser_confident && file_parts.size())
    {
        identified = true;
        title = file_parts.back();
    }
    // nothing left. maybe a remix album??
    else if (!file_parts.size() && immsdb->check_title(album))
    {
        identified = true;
        title = album;
    }
}
